/*
								+---------------------------------+
								|                                 |
								|   ***   S_Buffer engine   ***   |
								|                                 |
								|  Copyright   -tHE SWINe- 2002  |
								|                                 |
								|            Main.cpp             |
								|                                 |
								+---------------------------------+
*/

#include <windows.h>
#include <malloc.h>
#include <stdio.h>
#include <math.h>
#include "Directx.h"
#include "Matrix.h"
#include "Font_eng.h"
#include "Load_3ds.h"
#include "Raster.h"
#include "Main.h"


#define APPNAME "S_Buffer_Engine"
#define WNDNAME "S_Buffer - Engine 1.0"
static int Width = 640;
static int Height = 480;

unsigned __int32 *video;
unsigned __int32 *buffer;

#define PI 3.1415926

static World *svet;
// svet

static Matrix cmatrix;
static Matrix _no_y_cmatrix;
static float y_angle;
static int mx, my, ax, ay;
// kamera + kramy pro mouselook

static int nFrame = 0;
// frame counter (pocitadlo snimku)

static int have_to_graph;
// urcuje ze se ma vytvorit graph.txt - spousti se "Home"

// vytvori soubor engine.log
// a vlozi do nej uvodni radku
void Eng_Log_Init()
{
	FILE *fw;
	fw = fopen("Engine.log", "w");
	fprintf(fw, "__________BSP - Log__________\n");
	fclose(fw);
}

// pripoji k souboru engine.log text l_str a odradkuje
void Eng_Log(char *l_str)
{
	FILE *fw;
	fw = fopen("Engine.log", "a");
	fprintf(fw, "%s\n", l_str);
	fclose(fw);
}

// main
int WINAPI WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance, PSTR szCmdline, int nCmdShow)
{
	char szAppName[] = APPNAME;
	WNDCLASSEX wndclass;
	HWND hwnd;
	MSG msg;
	int i;

	Eng_Log_Init();

	wndclass.cbSize			= sizeof(wndclass);
	wndclass.style			= CS_HREDRAW | CS_VREDRAW;
	wndclass.lpfnWndProc	= Wokno;
	wndclass.cbClsExtra		= 0;
	wndclass.cbWndExtra		= 0;
	wndclass.hInstance		= hInstance;
	wndclass.hIcon			= LoadIcon(hInstance, IDI_APPLICATION);
	wndclass.hCursor		= LoadCursor(NULL, IDC_ARROW);
	wndclass.hbrBackground	= (HBRUSH)GetStockObject(BLACK_BRUSH);
	wndclass.lpszMenuName	= 0;
	wndclass.lpszClassName	= szAppName;
	wndclass.hIconSm		= LoadIcon(NULL, IDI_APPLICATION);
	RegisterClassEx(&wndclass);
	hwnd = CreateWindow(szAppName,
		WNDNAME, WS_DLGFRAME,
		CW_USEDEFAULT, CW_USEDEFAULT,
		Width, Height, NULL,
		NULL, hInstance, NULL);
	ShowWindow(hwnd, nCmdShow);
	UpdateWindow(hwnd);
	ShowCursor(FALSE);
	// vytvo okno

	Matrix_Init(&_no_y_cmatrix);
	Matrix_Rotate_Z(&_no_y_cmatrix, (float)-PI);
	Matrix_Translate(&_no_y_cmatrix, 0, -30, -50);
	y_angle = 0;
	// kamera ok

	Eng_Log("Loading world ..");

	if((svet = Load_3DS("Svet.3ds")) == NULL)
		return 0;
	for(i = 0; i < svet->n_object_num; i ++)
		Matrix_Init(&svet->object[i].m_object);
	// svet ok

	Eng_Log("    OK");

	Eng_Log("Loading textures ...");

	if(!Load_Textures(svet, "Textures"))
		return 0;
	// nahraje textury

	Eng_Log("    OK");

	Eng_Log("Init engine ...");
	
	Init_Engine(Width, Height);
	Init_Font_Engine(Width, Height);
	// engine ok

	Eng_Log("    OK");

	InitDirectX(hwnd, Width, Height);
	// inicializovano okno

	Eng_Log("DDraw is running..");

	Eng_Log("Locking some memory ...");

	if((buffer = (unsigned __int32*) malloc(Width * Height * 4)) == NULL)
		return 0;

	Eng_Log("    OK");

	Eng_Log("Rendering !");

	SetTimer(hwnd, 0, 20, NULL);

	SetCursorPos(Width >> 1, Height >> 1);
	mx = Width >> 1;
	my = Height >> 1;
	// pripravi MouseLook

	while (GetMessage (&msg, NULL, 0, 0)) {
		TranslateMessage (&msg);
		DispatchMessage (&msg);
	}

	free((void*)buffer);
	// uvolnuje pamet
	Terminate_Engine();
	Terminate_Font_Engine();
	// uvolnuje pamet v jinych modulech

	return msg.wParam;
}

// ati technologies filtr quincunx
void Quincunx(unsigned __int32 *src, unsigned __int32 *dest)
{
	int w2, h2, w1;
	int x, y;

	w2 = Width - 1;
	h2 = Height - 1;
	w1 = w2 + 1;

	y = Width * Height;
	for(x = 0; x < y; x ++)
		src[x] = ((src[x] >> 2) & 0x3f3f3f);

	for(y = 0; y < h2; y ++, dest += Width, src += Width) {
		for(x = 0; x < w2; x ++)
			dest[x] = src[x] + src[x + 1] + src[x + w1] + src[x + w2];
		dest[x] = src[x];
	}
	memcpy(dest, src, Width << 2);

	// obrazovy filtr
}

// funkce, zpracovavajici zpravy
// pri sipkach otaci kamerou, ins/del
// hybe kamerou dopredu dozadu
// a kresli obraz
// pri esc. skonci
LRESULT CALLBACK Wokno (HWND hwnd, UINT iMsg, WPARAM wParam, LPARAM lParam)
{
	float speeds[] = {2, 1, 0.5, 3};
	static int i;
	static int j = 0;
	char str[256];
	static float a = 0;
	POINT p;

	switch (iMsg) {
	case WM_KEYDOWN :
		if(wParam == 27) {
			ShowCursor(TRUE);
			SendMessage(hwnd, WM_CLOSE, 0, 0);
			return 0;
		} else {
			switch (wParam){
			case VK_LEFT:
				Matrix_Translate(&_no_y_cmatrix, -5, 0, 0);
				break;
			case VK_RIGHT:
				Matrix_Translate(&_no_y_cmatrix, 5, 0, 0);
				break;
			case VK_UP:
				Matrix_Translate(&_no_y_cmatrix, 0, -5 *
					(float)sin(y_angle), 5 * (float)cos(y_angle));
				break;
			case VK_DOWN:
				Matrix_Translate(&_no_y_cmatrix, 0, 5 *
					(float)sin(y_angle), -5 * (float)cos(y_angle));
				// ty kosinusy a sinusy zpusobi ze kamera se muze hybat i nahoru / dolu
				// tudiz uz je to presne jako kdyz v Half-Life nebo kdekoliv date "noclip"
				// (pokud by se rotovala jedinna matice kamery okolo osy x, tak by se mohla
				// otacet v ose pohledu, takze za chvili je vzhuru nohama - pokud smejdite
				// mysi porad dokolecka - a to nechci(byt vzhuru nohama))
				break;
			case VK_HOME:
				have_to_graph = 1;
				break;
			default:
				break;
			}
		} // posouvani kamery sipkama
		return 0;

	case WM_TIMER :
		GetCursorPos(&p);
		SetCursorPos(Width >> 1, Height >> 1);
		ax = p.x - mx;
		ay = p.y - my;
		Matrix_Rotate_Y(&_no_y_cmatrix, (float)ax * (float)3/900);
		y_angle -= (float)ay * (float)3/900;
		memcpy(&cmatrix, &_no_y_cmatrix, sizeof(Matrix));
		Matrix_Rotate_X(&cmatrix, y_angle);
		// osetri mouselook

		memset(buffer, 0, 4 * Width * Height);
		// vycisti obrazovku

		a += (float)0.1;
		if(a > 6.28)
			a = 0;

		Init_Screen();

		for(i = 0; i < svet->n_object_num; i ++) {
			DrawObject(&svet->object[i].m_object, &cmatrix,
				&svet->object[i], svet->material);
		}
		// kresli svet

		Release_Screen(buffer);

		if(have_to_graph) {
			Generate_Scanline_Graph();
			have_to_graph = 0;
		}

		if(InitVideoBuffer(&video)) {
			///
			memcpy(video, buffer, 4 * Width * Height);
			//Quincunx(buffer, video);
			/// zkopruje obraz

			if(j % 10 == 1)
				sprintf(str, "%d.st Frame", j ++);
			else if(j % 10 == 2)
				sprintf(str, "%d.nd Frame", j ++);
			else
				sprintf(str, "%d.th Frame", j ++);
			String_Out_AddTr(str, 10, 10, video);
			// vypise cislo snimku
		
			FlipVideoBuffer();
		}
		return 0;

	case WM_DESTROY :
		TerminateDirectX();
		PostQuitMessage (0);
		return 0;
	}

	return DefWindowProc (hwnd, iMsg, wParam, lParam) ;
}





